import torch
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset
import mlflow
from typing import Any, Dict, Tuple
from ddi_fw.ml.evaluation_helper import Metrics, evaluate
from ddi_fw.ml.wrappers.model_wrapper import ModelWrapper
import ddi_fw.utils as utils
import numpy as np
 
class PTModelWrapper(ModelWrapper):
    def __init__(self, date, descriptor, model_func, **kwargs):
        super().__init__(date, descriptor, model_func, **kwargs)
        self.batch_size = kwargs.get('batch_size',128)
        self.epochs = kwargs.get('epochs',100)
        self.model_func = model_func
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        
        # self.optimizer = kwargs['optimizer']
        
        # self.criterion = kwargs['criterion']
        self.loss_function = torch.nn.CrossEntropyLoss()

    def fit_model(self, model , X_train, y_train, X_valid, y_valid):

       
        criterion = self.loss_function.to(self.device)
        optimizer =torch.optim.Adam(model.parameters(), lr=0.001)
        # optimizer = torch.optim.SGD(model.parameters(), lr=0.01, momentum=0.9)
        train_dataset = TensorDataset(torch.tensor(X_train, dtype=torch.float32), torch.tensor(y_train, dtype=torch.float32))
        valid_dataset = TensorDataset(torch.tensor(X_valid, dtype=torch.float32), torch.tensor(y_valid, dtype=torch.float32))
        train_loader = DataLoader(train_dataset, batch_size=self.batch_size, shuffle=False)
        valid_loader = DataLoader(valid_dataset, batch_size=self.batch_size, shuffle=False)

        best_loss = float('inf')
        best_model = None

        for epoch in range(self.epochs):
            model.train()
            for batch_X, batch_y in train_loader:
                batch_X = batch_X.to(self.device, non_blocking=True)
                batch_y = batch_y.to(self.device, non_blocking=True)
                optimizer.zero_grad()
                output = model(batch_X)
                loss = criterion(output, batch_y)
                loss.backward()
                optimizer.step()

            valid_loss = self._validate(model, criterion, valid_loader)
            print(f'Epoch {epoch+1}/{self.epochs}, Validation Loss: {valid_loss:.4f}')
            if valid_loss < best_loss:
                best_loss = valid_loss
                best_model = model.state_dict()

        model.load_state_dict(best_model)
        return model, best_loss

    def _validate(self, model, criterion, valid_loader):
        model.eval()
        total_loss = 0
        with torch.no_grad():
            for batch_X, batch_y in valid_loader:
                batch_X = batch_X.to(self.device, non_blocking=True)
                batch_y = batch_y.to(self.device, non_blocking=True)
                output = model(batch_X)
                loss = criterion(output, batch_y)
                total_loss += loss.item()
        return total_loss / len(valid_loader)

    def fit(self):
        models = {}
        models_val_acc = {}
        
        self.kwargs['input_shape'] = self.train_data.shape
        self.num_classes = len(np.unique(self.test_label, axis=0))
        self.kwargs['num_classes'] = self.num_classes
        
        for i, (train_idx, val_idx) in enumerate(zip(self.train_idx_arr, self.val_idx_arr)):
            print(f"Validation {i}")
            model = self.model_func(**self.kwargs).to(self.device)
            with mlflow.start_run(run_name=f'Validation {i}', description='CV models', nested=True) as cv_fit:
                X_train_cv = self.train_data[train_idx]
                y_train_cv = self.train_label[train_idx]
                X_valid_cv = self.train_data[val_idx]
                y_valid_cv = self.train_label[val_idx]
                model, best_loss = self.fit_model(model, X_train_cv, y_train_cv, X_valid_cv, y_valid_cv)
                models[f'{self.descriptor}_validation_{i}'] = model
                models_val_acc[f'{self.descriptor}_validation_{i}'] = best_loss

        best_model_key = min(models_val_acc,  key=lambda k: models_val_acc[k])
        best_model = models[best_model_key]
        return best_model, best_model_key

    def predict(self):
        test_dataset = TensorDataset(torch.tensor(self.test_data, dtype=torch.float32), torch.tensor(self.test_label, dtype=torch.float32))
        test_loader = DataLoader(test_dataset, batch_size=self.batch_size, shuffle=False)
        self.best_model.eval()
        preds = []
        with torch.no_grad():
            for batch_X, _ in test_loader:
                batch_X = batch_X.to(self.device, non_blocking=True)
                output = self.best_model(batch_X)
                preds.append(output)
        return torch.cat(preds, dim=0).cpu().numpy()

    def fit_and_evaluate(self) -> Tuple[Dict[str, Any], Metrics, Any]:
        with mlflow.start_run(run_name=self.descriptor, description="***", nested=True) as run:
            print(run.info.artifact_uri)
            best_model, best_model_key = self.fit()
            print(best_model_key)
            self.best_model = best_model
            pred = self.predict()
            pred = utils.convert_to_categorical(pred.astype(np.int32), self.num_classes)
            logs, metrics = evaluate(actual=self.test_label, pred=pred, info=self.descriptor)
            metrics.format_float()
            mlflow.log_metrics(logs)
            mlflow.log_param('best_cv', best_model_key)
            utils.compress_and_save_data(metrics.__dict__, run.info.artifact_uri, f'{self.date}_metrics.gzip')
            mlflow.log_artifact(f'{run.info.artifact_uri}/{self.date}_metrics.gzip')

            return logs, metrics, pred

# from typing import Any
# import mlflow
# import torch
# from ddi_fw.ml.evaluation_helper import Metrics, evaluate
# from ddi_fw.ml.model_wrapper import ModelWrapper


# class PTModelWrapper(ModelWrapper):
#     def __init__(self, date, descriptor, model_func, batch_size=128, epochs=100, **kwargs):
#         super().__init__(date, descriptor, model_func, batch_size, epochs)
#         self.optimizer = kwargs['optimizer']
#         self.criterion = kwargs['criterion']

#     def _create_dataloader(self, data, labels):
#         dataset = torch.utils.data.TensorDataset(data, labels)
#         return torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, shuffle=True)

#     def predict(self):
#         print(self.train_data.shape)

#         with mlflow.start_run(run_name=self.descriptor, description="***", nested=True) as run:
#             models = {}
#             # models_val_acc = {}

#             for i, (train_idx, val_idx) in enumerate(zip(self.train_idx_arr, self.val_idx_arr)):
#                 print(f"Validation {i}")

#                 with mlflow.start_run(run_name=f'Validation {i}', description='CV models', nested=True) as cv_fit:
#                     model = self.model_func(self.train_data.shape[1])
#                     models[f'validation_{i}'] = model

#                     # Create DataLoaders
#                     X_train_cv = torch.tensor(self.train_data[train_idx], dtype=torch.float16)
#                     y_train_cv = torch.tensor(self.train_label[train_idx], dtype=torch.float16)
#                     X_valid_cv = torch.tensor(self.train_data[val_idx], dtype=torch.float16)
#                     y_valid_cv = torch.tensor(self.train_label[val_idx], dtype=torch.float16)

#                     train_loader = self._create_dataloader(X_train_cv, y_train_cv)
#                     valid_loader = self._create_dataloader(X_valid_cv, y_valid_cv)

#                     optimizer = self.optimizer
#                     criterion = self.criterion
#                     best_val_loss = float('inf')

#                     for epoch in range(self.epochs):
#                         model.train()
#                         for batch_X, batch_y in train_loader:
#                             optimizer.zero_grad()
#                             output = model(batch_X)
#                             loss = criterion(output, batch_y)
#                             loss.backward()
#                             optimizer.step()

#                         model.eval()
#                         with torch.no_grad():
#                             val_loss = self._validate(model, valid_loader)

#                         # Callbacks after each epoch
#                         for callback in self.callbacks:
#                             callback.on_epoch_end(epoch, logs={'loss': loss.item(), 'val_loss': val_loss.item()})

#                         if val_loss < best_val_loss:
#                             best_val_loss = val_loss
#                             best_model = model

#                     # Evaluate on test data
#                     with torch.no_grad():
#                         pred = best_model(torch.tensor(self.test_data, dtype=torch.float16))
#                         logs, metrics = evaluate(
#                             actual=self.test_label, pred=pred.numpy(), info=self.descriptor)
#                         mlflow.log_metrics(logs)

#             return logs, metrics, pred.numpy()

#     def _validate(self, model, valid_loader):
#         total_loss = 0
#         criterion = self.criterion

#         for batch_X, batch_y in valid_loader:
#             output = model(batch_X)
#             loss = criterion(output, batch_y)
#             total_loss += loss.item()

#         return total_loss / len(valid_loader)
    
#     def fit_and_evaluate(self)  -> tuple[dict[str, Any], Metrics, Any]:
#         return None,None,None


from ..langchain.embeddings import PoolingStrategy,SumPoolingStrategy,MeanPoolingStrategy,SentenceTransformerDecorator,PretrainedEmbeddings,SBertEmbeddings,VLLMEmbeddings, ClinicalBERTEmbeddings, BioSentVecEmbeddings
from .sentence_splitter import SentenceSplitter,PassthroughTextSplitter
# from .storage import DataFrameToVectorDB, generate_embeddings
from .faiss_storage import BaseVectorStoreManager, FaissVectorStoreManager,generate_embeddings
from .chroma_storage import ChromaVectorStoreManager


from .ml_helper import MultiModalRunner
from .result import Result
from .wrappers import ModelWrapper, TFModelWrapper, PTModelWrapper,CatBoostModelWrapper,LogisticRegressionModelWrapper,MultinomialNBModelWrapper,XGBoostModelWrapper,RandomForestModelWrapper
from .evaluation_helper import evaluate
from .tracking_service import TrackingService
from .ensemble import EnsembleStrategy, VotingStrategy, AveragingStrategy, StackingStrategy, GenericEnsembleWrapper


from typing import Any, Dict, List, Literal, Union
import numpy as np
from pydantic import BaseModel, Field
from sklearn import metrics
from sklearn.metrics import accuracy_score, precision_recall_curve
from sklearn.metrics import f1_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import auc
from sklearn.metrics import classification_report
from sklearn.preprocessing import OneHotEncoder


def __format__(d: Union[Dict[str, Union[List[float], float]], float], floating_number_precision=4) -> Union[Dict[str, Union[List[float], float]], float]:
    if isinstance(d, dict):
        d = {k: __round__(v, floating_number_precision) for k, v in d.items()}
    else:
        d = round(d, floating_number_precision)
    return d


def __round__(v, floating_number_precision=4) -> Union[List[float], float]:
    if type(v) is list or type(v) is set:
        return [round(item, floating_number_precision) for item in v]
    else:
        return round(v, floating_number_precision)


class Metrics(BaseModel):
    label: str
    time: float = 0.0
    accuracy: float = 0.0
    precision: Any = None
    recall: Any = None
    f1_score: Any = None
    roc_auc: Any = None
    roc_aupr: Any = None
    classification_report: Any = None

    def set_classification_report(self, classification_report):
        self.classification_report = classification_report
        
    def set_time(self, time):
        self.time = time

    def set_accuracy(self, accuracy):
        self.accuracy = accuracy

    def set_precision(self, precision):
        self.precision = precision

    def set_recall(self, recall):
        self.recall = recall

    def set_f1_score(self, f1_score):
        self.f1_score = f1_score

    def set_roc_auc(self, roc_auc):
        self.roc_auc = roc_auc

    def set_roc_aupr(self, roc_aupr):
        self.roc_aupr = roc_aupr

    def format_float(self, floating_number_precision=4):
        self.time = round(self.time, floating_number_precision)
        self.accuracy = round(self.accuracy, floating_number_precision)
        self.precision = __format__(self.precision, floating_number_precision)
        self.recall = __format__(self.recall, floating_number_precision)
        self.f1_score = __format__(self.f1_score, floating_number_precision)
        self.roc_auc = __format__(self.roc_auc, floating_number_precision)
        self.roc_aupr = __format__(self.roc_aupr, floating_number_precision)


# taken from https://github.com/YifanDengWHU/DDIMDL/blob/master/DDIMDL.py#L214
def roc_aupr_score(y_true, y_score, average="macro"):
    def _binary_roc_aupr_score(y_true, y_score):
        precision, recall, pr_thresholds = precision_recall_curve(
            y_true, y_score)
        # precision, recall, pr_thresholds = metrics.roc_curve(y, pred, pos_label=2)
        return auc(precision, recall)

    def _average_binary_score(binary_metric, y_true, y_score, average):  # y_true= y_one_hot
        if average == "binary":
            return binary_metric(y_true, y_score)
        if average == "micro":
            y_true = y_true.ravel()
            y_score = y_score.ravel()
        if y_true.ndim == 1:
            y_true = y_true.reshape((-1, 1))
        if y_score.ndim == 1:
            y_score = y_score.reshape((-1, 1))
        n_classes = y_score.shape[1]
        score = np.zeros((n_classes,))
        for c in range(n_classes):
            y_true_c = y_true.take([c], axis=1).ravel()
            y_score_c = y_score.take([c], axis=1).ravel()
            score[c] = binary_metric(y_true_c, y_score_c)
        return np.average(score)

    return _average_binary_score(_binary_roc_aupr_score, y_true, y_score, average)


def evaluate(actual: np.ndarray, pred: np.ndarray, info='', print_detail=False):
    '''actual and pred are one-hot encoded vectors'''
    y_true = actual
    y_pred = pred

    # Generate classification report
    c_report = classification_report(y_true, y_pred, output_dict=True)

    # Metrics initialization
    metrics = Metrics(label=info)

    n_classes = actual.shape[1]
    # n_classes = len(np.unique(actual))

    precision = {}
    recall = {}
    f_score = {}
    roc_aupr = {}
    roc_auc = {
        "weighted": 0.0,
        "macro": 0.0,
        "micro": 0.0
    }

    # Preallocate lists
    precision_vals: List[np.ndarray] = [np.array([]) for _ in range(n_classes)]
    recall_vals: List[np.ndarray] = [np.array([]) for _ in range(n_classes)]

    # Compute metrics for each class
    for i in range(n_classes):
        precision_vals[i], recall_vals[i], _ = precision_recall_curve(
            actual[:, i], pred[:, i])
        roc_aupr[i] = auc(recall_vals[i], precision_vals[i])

    # Calculate ROC AUC scores
    roc_auc["weighted"] = float(roc_auc_score(
        actual, pred, multi_class='ovr', average='weighted'))
    roc_auc["macro"] = float(roc_auc_score(
        actual, pred, multi_class='ovr', average='macro'))
    roc_auc["micro"] = float(roc_auc_score(
        actual, pred, multi_class='ovr', average='micro'))

    # Micro-average Precision-Recall curve and ROC-AUPR
    precision["micro_event"], recall["micro_event"], _ = precision_recall_curve(
        actual.ravel(), pred.ravel())
    roc_aupr["micro"] = auc(recall["micro_event"], precision["micro_event"])

    # Convert lists to numpy arrays for better performance
    precision["micro_event"] = precision["micro_event"].tolist()
    recall["micro_event"] = recall["micro_event"].tolist()

    # Overall accuracy
    acc = accuracy_score(y_true, y_pred)

    # Aggregate precision, recall, and f_score
    # for avg_type in ['weighted', 'macro', 'micro']:
    for avg_type in Literal['weighted', 'macro', 'micro'].__args__:
        precision[avg_type] = precision_score(y_true, y_pred, average=avg_type)
        recall[avg_type] = recall_score(y_true, y_pred, average=avg_type)
        f_score[avg_type] = f1_score(y_true, y_pred, average=avg_type)

    if print_detail:
        print(
            f'''Accuracy: {acc}
            , Precision:{precision['weighted']}
            , Recall: {recall['weighted']}
            , F1-score: {f_score['weighted']}
            ''')

    logs = {'accuracy': acc,
            'weighted_precision': precision['weighted'],
            'macro_precision': precision['macro'],
            'micro_precision': precision['micro'],
            'weighted_recall_score': recall['weighted'],
            'macro_recall_score': recall['macro'],
            'micro_recall_score': recall['micro'],
            'weighted_f1_score': f_score['weighted'],
            'macro_f1_score': f_score['macro'],
            'micro_f1_score': f_score['micro'],
            # 'weighted_roc_auc_score': weighted_roc_auc_score,
            # 'macro_roc_auc_score': macro_roc_auc_score,
            # 'micro_roc_auc_score': micro_roc_auc_score,
            # 'macro_aupr_score': macro_aupr_score,
            # 'micro_aupr_score': micro_aupr_score
            "micro_roc_aupr": roc_aupr['micro'],
            # "micro_precision_from_precision_recall_curve":precision["micro"],
            # "micro_recall_from_precision_recall_curve":recall["micro"],
            "weighted_roc_auc": roc_auc['weighted'],
            "macro_roc_auc": roc_auc['macro'],
            "micro_roc_auc": roc_auc['micro']
            }
    metrics.set_accuracy(acc)
    metrics.set_precision(precision)
    metrics.set_recall(recall)
    metrics.set_f1_score(f_score)
    metrics.set_roc_auc(roc_auc)
    metrics.set_roc_aupr(roc_aupr)
    metrics.set_classification_report(c_report)
    return logs, metrics


def evaluate_ex(actual, pred, info='', print_detail=False):
    # Precompute y_true and y_pred
    y_true = np.argmax(actual, axis=1)
    y_pred = np.argmax(pred, axis=1)

    # Generate classification report
    c_report = classification_report(y_true, y_pred, output_dict=True)

    # Metrics initialization
    metrics = Metrics(label=info)

    n_classes = actual.shape[1]

    precision = {}
    recall = {}
    f_score = {}
    roc_aupr = {}
    roc_auc = {
        "weighted": 0.0,
        "macro": 0.0,
        "micro": 0.0
    }

    # Preallocate lists
    precision_vals: List[np.ndarray] = [np.array([]) for _ in range(n_classes)]
    recall_vals: List[np.ndarray] = [np.array([]) for _ in range(n_classes)]

    # Compute metrics for each class
    for i in range(n_classes):
        precision_vals[i], recall_vals[i], _ = precision_recall_curve(
            actual[:, i], pred[:, i])
        roc_aupr[i] = auc(recall_vals[i], precision_vals[i])

    # Calculate ROC AUC scores
    roc_auc["weighted"] = float(roc_auc_score(
        actual, pred, multi_class='ovr', average='weighted'))
    roc_auc["macro"] = float(roc_auc_score(
        actual, pred, multi_class='ovr', average='macro'))
    roc_auc["micro"] = float(roc_auc_score(
        actual, pred, multi_class='ovr', average='micro'))

    # Micro-average Precision-Recall curve and ROC-AUPR
    precision["micro_event"], recall["micro_event"], _ = precision_recall_curve(
        actual.ravel(), pred.ravel())
    roc_aupr["micro"] = auc(recall["micro_event"], precision["micro_event"])

    # Convert lists to numpy arrays for better performance
    precision["micro_event"] = precision["micro_event"].tolist()
    recall["micro_event"] = recall["micro_event"].tolist()

    # Overall accuracy
    acc = accuracy_score(y_true, y_pred)

    # Aggregate precision, recall, and f_score
    # for avg_type in ['weighted', 'macro', 'micro']:
    for avg_type in Literal['weighted', 'macro', 'micro'].__args__:
        precision[avg_type] = precision_score(y_true, y_pred, average=avg_type)
        recall[avg_type] = recall_score(y_true, y_pred, average=avg_type)
        f_score[avg_type] = f1_score(y_true, y_pred, average=avg_type)

    if print_detail:
        print(
            f'''Accuracy: {acc}
            , Precision:{precision['weighted']}
            , Recall: {recall['weighted']}
            , F1-score: {f_score['weighted']}
            ''')

    logs = {'accuracy': acc,
            'weighted_precision': precision['weighted'],
            'macro_precision': precision['macro'],
            'micro_precision': precision['micro'],
            'weighted_recall_score': recall['weighted'],
            'macro_recall_score': recall['macro'],
            'micro_recall_score': recall['micro'],
            'weighted_f1_score': f_score['weighted'],
            'macro_f1_score': f_score['macro'],
            'micro_f1_score': f_score['micro'],
            # 'weighted_roc_auc_score': weighted_roc_auc_score,
            # 'macro_roc_auc_score': macro_roc_auc_score,
            # 'micro_roc_auc_score': micro_roc_auc_score,
            # 'macro_aupr_score': macro_aupr_score,
            # 'micro_aupr_score': micro_aupr_score
            "micro_roc_aupr": roc_aupr['micro'],
            # "micro_precision_from_precision_recall_curve":precision["micro"],
            # "micro_recall_from_precision_recall_curve":recall["micro"],
            "weighted_roc_auc": roc_auc['weighted'],
            "macro_roc_auc": roc_auc['macro'],
            "micro_roc_auc": roc_auc['micro']
            }
    metrics.set_accuracy(acc)
    metrics.set_precision(precision)
    metrics.set_recall(recall)
    metrics.set_f1_score(f_score)
    metrics.set_roc_auc(roc_auc)
    metrics.set_roc_aupr(roc_aupr)
    metrics.set_classification_report(c_report)
    return logs, metrics



from ddi_fw.ml.wrappers.model_wrapper import ModelWrapper
from ddi_fw.ml.wrappers.tensorflow_wrapper import TFModelWrapper
from ddi_fw.ml.wrappers.pytorch_wrapper import PTModelWrapper
from ddi_fw.ml.wrappers.catboost_wrapper import CatBoostModelWrapper
from ddi_fw.ml.wrappers.xgboost_wrapper import XGBoostModelWrapper
from ddi_fw.ml.wrappers.logistic_regression_wrapper import LogisticRegressionModelWrapper
from ddi_fw.ml.wrappers.mnb_wrapper import MultinomialNBModelWrapper
from ddi_fw.ml.wrappers.random_forest_wrapper import RandomForestModelWrapper

__all__ = ["ModelWrapper", "TFModelWrapper", "PTModelWrapper", "CatBoostModelWrapper", "XGBoostModelWrapper",
           "LogisticRegressionModelWrapper", "MultinomialNBModelWrapper", "RandomForestModelWrapper"]



from typing import Type
from ddi_fw.datasets import BaseDataset
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
import pandas as pd

from ddi_fw.datasets.dataset_splitter import DatasetSplitter


class IrisDataset(BaseDataset):
    dataset_name: str = "IrisDataset"
    columns: list[str] = ['sepal_length',
                          'sepal_width', 'petal_length', 'petal_width']
    index_path: str = "tests/datasets/iris/indexes"
    dataset_splitter_type: Type[DatasetSplitter] = DatasetSplitter

    def prep(self):
        print(self.columns)
        # Load the Iris dataset
        iris = load_iris()
        X = iris.data
        print(iris.target.shape)
        y = iris.target.reshape(-1, 1)
        print(y.shape)
        # # One-hot encode the target variable
        encoder = OneHotEncoder(sparse_output=False)
        y = encoder.fit_transform(y)
        replace_dict = {
            'sepal length (cm)': 'sepal_length',
            'sepal width (cm)': 'sepal_width',
            'petal length (cm)': 'petal_length',
            'petal width (cm)': 'petal_width'
        }

        # Replace the columns using the dictionary

        # # Create a DataFrame for the dataset
        self.dataframe = pd.DataFrame(
            X, columns=iris.feature_names, index=range(X.shape[0]))
        self.dataframe.rename(columns=replace_dict, inplace=True)
        self.dataframe['class'] = list(y)
        # self.dataframe['class'] = y.flatten()



from ddi_fw.ml.ensemble.ensemble_strategy import EnsembleStrategy, VotingStrategy, AveragingStrategy, StackingStrategy
from ddi_fw.ml.ensemble.ensemble_stacking_wrapper import StackingWrapper
from ddi_fw.ml.ensemble.ensemble_voting_wrapper import VotingWrapper
from ddi_fw.ml.ensemble.ensemble_wrapper import GenericEnsembleWrapper
__all__ = ["EnsembleStrategy", "VotingStrategy", "AveragingStrategy", "StackingStrategy", "GenericEnsembleWrapper", "StackingWrapper", "VotingWrapper"]



from typing import Any, Dict, List, Optional

import numpy as np
from ddi_fw import utils
from ddi_fw.ml.ensemble.ensemble_strategy import AveragingStrategy, EnsembleStrategy, StackingStrategy, VotingStrategy
from ddi_fw.ml.evaluation_helper import Metrics, evaluate
from ddi_fw.ml.result import Result
from ddi_fw.ml.wrappers.model_wrapper import ModelWrapper


class GenericEnsembleWrapper(ModelWrapper):
    """Generic ensemble wrapper supporting multiple strategies"""

    STRATEGY_MAP = {
        "voting": VotingStrategy,
        "averaging": AveragingStrategy,
        "stacking": StackingStrategy
    }

    def __init__(
        self,
        base_wrappers: List[ModelWrapper],
        meta_learners: List[ModelWrapper],
        meta_learner_processors: Dict[str, Any],
        ensemble_strategies: str,          # now comma separated input
        name: str = "ensemble",
        tracking_service: Optional[Any] = None,
        date: Optional[str] = None
    ):
        self.base_wrappers = base_wrappers
        self.name = name
        self.tracking_service = tracking_service
        self.date = date or utils.utc_time_as_string_simple_format()
        self.ensemble_strategies = ensemble_strategies
        self.meta_learners = meta_learners
        self.meta_learner_processors = meta_learner_processors

        # prediction cache so we don't recompute
        self.wrapper_predictions: Dict[str, np.ndarray] = {}

        self.individual_metrics = {}
        self.ensemble_metrics: Dict[str, Metrics] = {}
        self.result = Result()

    # ---------------------------------------------------------
    # FIT
    # ---------------------------------------------------------
    def fit(self):
        train_label, test_label = None, None

        # if any(getattr(wrapper, 'best_model', None) is not None for wrapper in self.base_wrappers):
        #     # At least one wrapper has best_model set
        #     pass

        train_label, test_label = None, None
        train_idx_arr, val_idx_arr = None, None
        val_labels = []
        inputs = []

        for wrapper in self.base_wrappers:
            train_label = wrapper.train_label
            test_label = wrapper.test_label

            train_idx_arr = wrapper.train_idx_arr
            val_idx_arr = wrapper.val_idx_arr

            if not val_labels:
                for i, (train_idx, val_idx) in enumerate(zip(train_idx_arr, val_idx_arr)):
                    fold_val_labels = train_label[val_idx]
                    val_labels.append(fold_val_labels)
            # Convert the list of arrays to a single numpy array
            all_fold_val_labels = np.concatenate(val_labels)

            print(f"Training base model: {wrapper.descriptor}")
            best_model, _, val_preds_dict = wrapper.fit()
            num_classes = wrapper.num_classes
            wrapper.best_model = best_model
            # pred_train = wrapper.predict()
            # if val_preds_dict is not None:
            print("val_preds_dict")
            print(val_preds_dict)
            for key, value in val_preds_dict.items():
                print(f"Key: {key}, Value shape: {value.shape}")
            # stacked_values = np.concatenate(list(val_preds_dict.values()))
            # stacked_values = np.stack(list(val_preds_dict.values()), axis=1)
            stacked_values = np.vstack(list(val_preds_dict.values()))

            inputs.append(stacked_values)
            # inputs.append(list(val_preds_dict.values()))
      
        # inputs = np.concatenate(inputs, axis=1)
        print("shape of inputs for meta-learner:")
        print(inputs[0].shape)
        self.num_classes = num_classes
        self.test_label = test_label
        print("val labels")
        print(len(val_labels))
        # X_meta_train = np.stack(inputs, axis=1)
        # y_meta_train = train_label
        y_meta_train = all_fold_val_labels
        # print("X_meta_train shape:", X_meta_train.shape)
        print(y_meta_train)
        print("y_meta_train shape:", y_meta_train.shape)

        # Step 3: set the meta-learner's training data
        for meta_learner in self.meta_learners:
            
            # meta_learner.train_data = X_meta_train
            meta_learner.train_data = inputs
            meta_learner.train_label = y_meta_train
            meta_learner.train_idx_arr = train_idx_arr
            meta_learner.val_idx_arr = val_idx_arr
            meta_learner.test_label = test_label
            meta_learner.num_classes = num_classes

        # Parse comma-separated strategies
        self.selected_strategies: List[EnsembleStrategy] = []
        for s in self.ensemble_strategies.split(","):
            s = s.strip().lower()
            if s not in self.STRATEGY_MAP:
                raise ValueError(f"Unknown strategy '{s}'")
            if s == "stacking":
                if not self.meta_learners:
                    raise ValueError(
                        "Meta-learner must be provided for stacking strategy")
                for meta_learner in self.meta_learners:
                    processor_dict = self.meta_learner_processors.get(meta_learner.descriptor)
                    processor = processor_dict.get("processor") if processor_dict else None
                    processor_config = processor_dict.get("processor_config") if processor_dict else None
                    print("processor_dict for stacking:")
                    print(processor_dict)
                    if processor is None:
                        raise ValueError(
                            f"Processor must be provided for meta-learner '{meta_learner.descriptor}' in stacking strategy")
                    
                    if processor_config is None:
                        raise ValueError(
                            f"Processor config must be provided for meta-learner '{meta_learner.descriptor}' in stacking strategy")
                    
                    self.selected_strategies.append(StackingStrategy(descriptor=meta_learner.descriptor,
                                                                     meta_learner=meta_learner, 
                                                                     base_wrappers=self.base_wrappers,
                                                                     processor=processor,
                                                                     processor_config=processor_config))
            else:
                self.selected_strategies.append(self.STRATEGY_MAP[s]())

    # ---------------------------------------------------------
    # PREDICT — with prediction caching
    # ---------------------------------------------------------

    def _compute_and_cache_predictions(self):
        """Compute wrapper predictions only once."""
        if self.wrapper_predictions:
            return  # already computed

        for wrapper in self.base_wrappers:
            descriptor = wrapper.descriptor

            y_pred = wrapper.predict()
            print(f"Predictions from wrapper {descriptor}:")
            print(y_pred)
            print("Shape:")
            print(y_pred.shape)

            # Store raw predictions (strategies want raw data)
            self.wrapper_predictions[descriptor] = y_pred

            # Evaluation for individual wrapper
            y_pred_cat = utils.convert_to_categorical(y_pred, self.num_classes)
            _, wrapper_metrics = evaluate(self.test_label, y_pred_cat)
            wrapper_metrics.set_time(wrapper.elapsed_time)
            wrapper_metrics.format_float()

            self.result.add_metric(descriptor, wrapper_metrics)
            self.individual_metrics[descriptor] = wrapper_metrics
        print(self.wrapper_predictions)
    # ---------------------------------------------------------
    # MAIN PREDICTION PIPELINE
    # ---------------------------------------------------------

    def predict(self):
        # Compute predictions only once
        self._compute_and_cache_predictions()

        pred_list = list(self.wrapper_predictions.values())

        ensemble_outputs = {}

        for strategy in self.selected_strategies:
            strategy_name = strategy.get_strategy_name()

            # Combine predictions
            ensemble_pred = strategy.combine_predictions(pred_list)
            ensemble_pred = utils.convert_to_categorical(
                ensemble_pred, self.num_classes)

            # Evaluate ensemble
            _, metrics = evaluate(self.test_label, ensemble_pred)
            if type(strategy) is StackingStrategy:
                metrics.set_time(strategy.meta_learner.elapsed_time)

            metrics.format_float()

            # token = f"ensemble_{strategy_name}"
            token = strategy_name
            self.ensemble_metrics[token] = metrics
            self.result.add_metric(token, metrics)

            ensemble_outputs[token] = ensemble_pred

        return ensemble_outputs   # multiple ensemble predictions

    # ---------------------------------------------------------
    # FIT AND EVALUATE
    # ---------------------------------------------------------
    def fit_and_evaluate(self):
        self.fit()
        ensemble_preds = self.predict()
        return None, self.result, ensemble_preds



from collections import defaultdict
from typing import Any, Dict, List, Optional, Type
from ddi_fw.datasets.dataset_splitter import DatasetSplitter

from pydantic import BaseModel
from ddi_fw.datasets.core import TextDatasetMixin
from ddi_fw.langchain.faiss_storage import BaseVectorStoreManager
# from ddi_fw.ml.ensemble.ensemble_strategy import AveragingStrategy, StackingStrategy, VotingStrategy
from ddi_fw.ml import Result
from ddi_fw.ml.ensemble.ensemble_wrapper import GenericEnsembleWrapper
from ddi_fw.ml.tracking_service import TrackingService
from ddi_fw.langchain.embeddings import PoolingStrategy
from ddi_fw.datasets import BaseDataset
# from ddi_fw.ml import MultiModalRunner
import logging
from ddi_fw.ml.wrappers.pytorch_wrapper import PTModelWrapper
from ddi_fw.ml.wrappers.tensorflow_wrapper import TFModelWrapper
from ddi_fw.utils import utils
from ddi_fw.utils.package_helper import get_import


class EnsemblePipeline(BaseModel):

    library: str = 'tensorflow'
    ensemble_strategy: str = 'voting'  # 'voting', 'averaging', 'stacking'
    experiment_name: str
    experiment_description: str
    tracking_library: str
    tracking_params: Optional[Dict[str, Any]] = None
    dataset_type: Type[BaseDataset]
    dataset_additional_config: Optional[Dict[str, Any]] = None
    dataset_splitter_type: Type[DatasetSplitter] = DatasetSplitter
    columns: Optional[List[str]] = None
    input_processing: Optional[List[Dict[str, Any]]] = None
    embedding_dict: Optional[Dict[str, Any]] = None
    column_embedding_configs: Optional[List] = None
    vector_db_persist_directory: Optional[str] = None
    vector_db_collection_name: Optional[str] = None
    embedding_pooling_strategy_type: Type[PoolingStrategy] | None = None
    vector_store_manager_type: Type[BaseVectorStoreManager] | None = None
    combinations: Optional[List[tuple]] = None
    model: Optional[Any] = None
    default_model:  Optional[Any] = None
    multi_modal:  Optional[Any] = None
    _tracking_service: TrackingService | None = None
    _dataset: BaseDataset | None = None
    _items: List = []
    _train_idx_arr: List | None = []
    _val_idx_arr: List | None = []

    @property
    def tracking_service(self) -> TrackingService | None:
        return self._tracking_service

    @property
    def dataset(self) -> BaseDataset | None:
        return self._dataset

    @property
    def items(self) -> List:
        return self._items

    @property
    def train_idx_arr(self) -> List | None:
        return self._train_idx_arr

    @property
    def val_idx_arr(self) -> List | None:
        return self._val_idx_arr

    class Config:
        arbitrary_types_allowed = True

    # TODO embedding'leri set etme kimin görevi
    def build(self):
        self._tracking_service = TrackingService(self.experiment_name,
                                                 backend=self.tracking_library, tracking_params=self.tracking_params)

        if self.embedding_pooling_strategy_type is not None and not isinstance(self.embedding_pooling_strategy_type, type):
            raise TypeError(
                "self.embedding_pooling_strategy_type must be a class, not an instance")
        if not isinstance(self.dataset_type, type):
            raise TypeError(
                "self.dataset_type must be a class, not an instance")

        # 'enzyme','target','pathway','smile','all_text','indication', 'description','mechanism_of_action','pharmacodynamics', 'tui', 'cui', 'entities'
        kwargs = {"columns": self.columns,
                  "additional_config": self.dataset_additional_config}

        dataset_splitter = self.dataset_splitter_type()
        pooling_strategy = self.embedding_pooling_strategy_type(
        ) if self.embedding_pooling_strategy_type else None

        params = {}

        if self.embedding_dict is not None:
            params["embedding_dict"] = self.embedding_dict
        if self.vector_db_persist_directory is not None:
            params["persist_directory"] = self.vector_db_persist_directory
        if self.vector_db_collection_name is not None:
            params["collection_name"] = self.vector_db_collection_name

        vector_store_manager = self.vector_store_manager_type(
            **params) if self.vector_store_manager_type else None
        if issubclass(self.dataset_type, TextDatasetMixin):

            dataset = self.dataset_type(
                vector_store_manager=vector_store_manager,
                embedding_dict=self.embedding_dict,
                pooling_strategy=pooling_strategy,
                column_embedding_configs=self.column_embedding_configs,
                vector_db_persist_directory=self.vector_db_persist_directory,
                vector_db_collection_name=self.vector_db_collection_name,
                dataset_splitter_type=self.dataset_splitter_type,
                **kwargs)

        elif self.dataset_type == BaseDataset:  # !!! check it
            dataset = self.dataset_type(
                dataset_splitter_type=self.dataset_splitter_type,
                **kwargs)
        else:
            dataset = self.dataset_type(
                dataset_splitter_type=self.dataset_splitter_type, **kwargs)

        dataset.input_processing = self.input_processing

        # X_train, X_test, y_train, y_test, train_indexes, test_indexes, train_idx_arr, val_idx_arr = dataset.load()

        dataset.load()

        self._dataset = dataset
        self._train_idx_arr = dataset.train_idx_arr
        self._val_idx_arr = dataset.val_idx_arr

        dataframe = dataset.dataframe

        # Check if any of the arrays are None or empty
        is_data_valid = (dataset.X_train is not None and dataset.X_train.size > 0 and
                         dataset.y_train is not None and dataset.y_train.size > 0 and
                         dataset.X_test is not None and dataset.X_test.size > 0 and
                         dataset.y_test is not None and dataset.y_test.size > 0)

        # Check if the dataframe is None or empty
        is_dataframe_valid = dataframe is not None and not dataframe.empty

        if not (is_data_valid or is_dataframe_valid):
            raise ValueError("The dataset is not loaded")

        # column name, train data, train label, test data, test label
        self._items = dataset.produce_inputs()

        print(f"Building the experiment: {self.experiment_name}")
        # print(
        #     f"Name: {self.experiment_name}, Dataset: {dataset}, Model: {self.model}")
        # Implement additional build logic as needed
        return self

    def __create_wrapper(self, library):
        if library == 'tensorflow':
            return TFModelWrapper
        elif library == 'pytorch':
            return PTModelWrapper
        elif library =="catboost":
            from ddi_fw.ml.wrappers.catboost_wrapper import CatBoostModelWrapper
            return CatBoostModelWrapper
        elif library =="xgboost":
            from ddi_fw.ml.wrappers.xgboost_wrapper import XGBoostModelWrapper
            return XGBoostModelWrapper
        elif library =="logistic_regression":
            from ddi_fw.ml.wrappers.logistic_regression_wrapper import LogisticRegressionModelWrapper
            return LogisticRegressionModelWrapper
        elif library =="multinomial_nb":
            from ddi_fw.ml.wrappers.mnb_wrapper import MultinomialNBModelWrapper
            return MultinomialNBModelWrapper
        elif library =="random_forest":
            from ddi_fw.ml.wrappers.random_forest_wrapper import RandomForestModelWrapper
            return RandomForestModelWrapper
        try:
            lib = get_import(library)
            return lib
        except Exception:
            raise ValueError(
                "Unsupported library type. Supported types are 'tensorflow', 'pytorch' or custom model wrapper class path.")

    def run(self):
        if self.default_model is None:
            raise ValueError("default_model cannot be None")
        if self.multi_modal is None:
            raise ValueError("multi_modal cannot be None")
        for m in self.multi_modal:
            if 'ensemble_task' not in m:
                raise ValueError(
                    "Each multi_modal model must have an 'ensemble_task' defined ('base_model' or 'meta_learner').")
        if self._tracking_service is None:
            logging.warning("Tracking service is not initialized.")
        else:
            self._tracking_service.setup()

        date = utils.utc_time_as_string_simple_format()

        meta_learner_wrappers = []
        base_model_wrappers = []
        # Separate base models and meta learner models
        meta_learner_models = [
            m for m in self.multi_modal if m['ensemble_task'] == 'meta_learner']
        base_models = [
            m for m in self.multi_modal if m['ensemble_task'] == 'base_model']
        if base_models:
            for m in base_models:
                name = m.get('name')
                library = m.get('library', self.library)
                model_type = get_import(m.get("model_type"))
                kwargs = m.get('params')

                input = m.get('input')
                inputs = m.get('inputs')

                if model_type is None:
                    raise Exception(
                        "model_type cannot be None, it should be defined in multi_modal or default_model")

                if input and inputs:
                    raise ValueError(
                        "Only one of 'input' or 'inputs' should be defined.")
                if not input and not inputs:
                    raise ValueError(
                        "At least one of 'input' or 'inputs' must be defined.")

                if input and not isinstance(input, str):
                    raise ValueError(
                        "'input' should be a single string. For multiple inputs, use 'inputs'.")

                    # Get stacking and reshaping config
                processor_type = m.get(
                    "processor", "ddi_fw.datasets.processor.DefaultInputProcessor")
                # Ensure the processor type is valid
                processor = get_import(processor_type)
                force_stack = m.get("force_stack", True)
                reshape_dims = m.get("reshape")
                train_data, train_label, test_data, test_label = None, None, None, None

                # Prepare processing config with all context
                processing_config = {
                    "force_stack": force_stack,
                    "reshape": reshape_dims
                }
                item_dict = {t[0]: t for t in self.items}
                # --- SINGLE INPUT CASE ---
                if input:
                    item = item_dict[input]
                    train_data = item[1]
                    train_label = item[2]
                    test_data = item[3]
                    test_label = item[4]
                elif inputs:
                    filtered_dict = {k: item_dict[k]
                                     for k in inputs if k in item_dict}
                    if not filtered_dict:
                        raise ValueError(
                            f"No matching inputs found in item_dict for: {inputs}")

                    first_input = next(iter(filtered_dict.values()))
                    train_data = [f[1] for f in filtered_dict.values()]
                    test_data = [f[3] for f in filtered_dict.values()]
                    train_label = first_input[2]
                    test_label = first_input[4]

                train_data = processor().process2(train_data, processing_config)
                test_data = processor().process2(test_data, processing_config)

                T = self.__create_wrapper(library)
                wrapper = T(date, name, model_type,
                            tracking_service=self.tracking_service,  **kwargs)
                wrapper.set_data(
                    self.train_idx_arr, self.val_idx_arr,
                    train_data, train_label,
                    test_data, test_label
                )
                base_model_wrappers.append(wrapper)

        # Get ensemble strategy from config (default: voting)
            ensemble_strategies = self.ensemble_strategy  # e.g. "voting,stacking"

            # Prepare meta learner only if stacking is included
            meta_learner_wrapper = None
            meta_learner_processors = defaultdict(dict)
            if "stacking" in ensemble_strategies.replace(" ", "").lower():

                for m in meta_learner_models:
                    name = m.get('name')
                    library = m.get('library', self.library)
                    model_type = get_import(m.get("model_type"))
                    meta_kwargs = m.get('params')

                    processor_type = m.get(
                        "processor", "ddi_fw.datasets.processor.DefaultInputProcessor")
                    # Ensure the processor type is valid
                    processor = get_import(processor_type)
                    force_stack = m.get("force_stack", True)
                    reshape_dims = m.get("reshape", None)
                    reduce_dimensions = m.get("reduce_dimensions",False)
                    flatten = m.get("flatten", False)

                    # Prepare processing config with all context
                    processing_config = {
                        "force_stack": force_stack,
                        "reshape": reshape_dims,
                        "reduce_dimensions": reduce_dimensions,
                        "flatten": flatten
                    }

                    meta_learner_processors[name] = {"processor": processor, "processor_config": processing_config}
                    T_meta = self.__create_wrapper(library)
                    meta_learner_wrapper = T_meta(
                        date,
                        name,
                        model_type,
                        tracking_service=self.tracking_service,
                        **meta_kwargs
                    )
                    meta_learner_wrappers.append(meta_learner_wrapper)

            # Instantiate ensemble wrapper (supports multiple strategies)
            ensemble = GenericEnsembleWrapper(
                base_wrappers=base_model_wrappers,
                meta_learners=meta_learner_wrappers,
                meta_learner_processors = meta_learner_processors,
                ensemble_strategies=ensemble_strategies,
                name="ensemble",
                tracking_service=self.tracking_service,
                date=date
            )

            # Train and evaluate
            logs, result, pred = ensemble.fit_and_evaluate()

            # result = Result()
            # result.add_metric(ensemble.name, metrics)
            return result



import abc
from collections import defaultdict
import glob
import logging
from typing import Any, Dict, List, Optional, Type
# import chromadb
# from chromadb.api.types import IncludeEnum
import numpy as np
import pandas as pd
from pydantic import BaseModel, Field, computed_field
from ddi_fw.datasets.dataset_splitter import DatasetSplitter
from ddi_fw.langchain.faiss_storage import BaseVectorStoreManager
from ddi_fw.utils.utils import create_folder_if_not_exists
from sklearn.metrics.pairwise import cosine_similarity


try:
    from ddi_fw.vectorization import SimilarityMatrixGenerator, VectorGenerator
except ImportError:
    raise ImportError(
        "Failed to import vectorization module. Ensure that the module exists and is correctly installed. ")

try:
    from ddi_fw.langchain.embeddings import PoolingStrategy
except ImportError:
    raise ImportError(
        "Failed to import langchain.embeddings module. ")


def stack(df_column):
    return np.stack(df_column.values)


def generate_vectors(df, columns):
    vectorGenerator = VectorGenerator(df)
    generated_vectors = vectorGenerator.generate_feature_vectors(
        columns)
    return generated_vectors


def generate_sim_matrices_new(df, generated_vectors, columns, similarity_type='jaccard', key_column="id"):
    """
    Generate similarity matrices for given columns.

    similarity_type: either a string ('jaccard' or 'cosine') or a callable that accepts a matrix and returns
    a square similarity matrix (numpy array / array-like). Strings are mapped to methods on SimilarityMatrixGenerator.
    """
    sim_matrix_gen = SimilarityMatrixGenerator()

    # registry of built-in similarity names -> callable
    registry = {
        'jaccard': sim_matrix_gen.create_jaccard_similarity_matrices,
        'cosine': sim_matrix_gen.create_cosine_similarity_matrices,
    }

    # resolve similarity function
    if callable(similarity_type):
        sim_func = similarity_type
    elif isinstance(similarity_type, str):
        sim_func = registry.get(similarity_type.lower())
        if sim_func is None:
            raise ValueError(f"Unknown similarity_type '{similarity_type}'. Available: {list(registry.keys())}")
    else:
        raise TypeError("similarity_type must be a string or a callable")

    sim_dict = {}
    for column in columns:
        key = column
        matrix = generated_vectors[key]
        sim_dict[column] = sim_func(matrix)

    # convert to DataFrame with original keys as row/col labels
    similarity_matrices = {}
    keys = df[key_column].to_list()
    new_columns = {idx: keys[idx] for idx in range(len(keys))}

    for column in columns:
        new_df = pd.DataFrame.from_dict(sim_dict[column])
        new_df = new_df.rename(index=new_columns, columns=new_columns)
        similarity_matrices[column] = new_df

    return similarity_matrices

class BaseDataset(BaseModel, abc.ABC):
    dataset_name: str
    index_path: Optional[str] = None
    dataset_splitter_type: Type[DatasetSplitter]
    class_column: str = 'class'
    dataframe: Optional[pd.DataFrame] = None
    X_train: Optional[np.ndarray] = None
    X_test: Optional[np.ndarray] = None
    y_train: Optional[np.ndarray] = None
    y_test: Optional[np.ndarray] = None
    train_indexes: Optional[pd.Index] = None
    test_indexes: Optional[pd.Index] = None
    train_idx_arr: Optional[List[np.ndarray]] = None
    val_idx_arr: Optional[List[np.ndarray]] = None
    columns: List[str] = []
    additional_config: Optional[Dict[str, Any]] = None
    input_processing: Optional[List[Dict[str, Any]]] = None

    class Config:
        arbitrary_types_allowed = True
        
    def process_input_data(self,data, processing_config=None):
        
        if not processing_config:
            return data
        if processing_config.get("flatten", False):
            print("Flattening data...")
            data = np.array(data).flatten()
            print(f"Data shape after flattening: {data.shape}")
        
        if processing_config.get("stack", False):
            print("Stacking data...")
            data = np.stack(data)
            print(f"Data shape after stacking: {data.shape}")
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        # if processing_config.get("flatten", False):
        #     data = np.stack(data.flatten().tolist())
        # Ensure we start with a NumPy array
       

        # Normalize input
        if processing_config.get("normalize", False):
            data = data.astype(np.float32)
            max_val = np.max(data)
            if max_val > 1:
                data /= max_val

        # Reshape input (for images etc.)
        if "reshape" in processing_config:
            try:
                target_shape = tuple(processing_config["reshape"])
                data = data.reshape((-1, *target_shape))
            except Exception as e:
                raise ValueError(f"Reshape failed for data with shape {data.shape}: {e}")


        return data

    # TODO columns yoksa tüm feature'lar alınıyor, bu pipeline'da nasıl yapılacak?
    # TODO processor sınıfı kullanılsın
    def produce_inputs(self):
        # Grouping the list by "column" key
        grouped_data = defaultdict(dict)

        if self.input_processing:
            for item in self.input_processing:
                grouped_data[item["column"]] = item
                
        items = []
        if self.X_train is None or self.X_test is None:
            raise Exception("There is no data to produce inputs")
        y_train_label, y_test_label = np.array(
            self.y_train), np.array(self.y_test)

        if self.columns is None or len(self.columns) == 0 or len(self.columns) == 1:
            # If no columns or only one column are provided, do not change the data
            # and use the entire dataset as a single input.
            column = self.columns[0] if self.columns else 'default'
            train_data, test_data = self.X_train[:, :], self.X_test[:, :]
            processing_config = grouped_data[column]
            train_data = self.process_input_data(train_data, processing_config)
            test_data = self.process_input_data(test_data, processing_config)
            # train_data,test_data = np.stack(train_data.flatten().tolist()), np.stack(test_data.flatten().tolist())
            items.append([f'{column}', np.nan_to_num(train_data),
                          y_train_label, np.nan_to_num(test_data), y_test_label])
        else:
            for index, column in enumerate(self.columns):
                processing_config = grouped_data[column]
                train_data, test_data = self.X_train[:,
                                                     index], self.X_test[:, index]
                #TODO üstteki satır ile alttaki tek satır olsun, tolist() ile numpy array'e çevrilmesin, numpy array zaten ama uyarı verdiği için böyle
                train_data = self.process_input_data(train_data, processing_config)
                test_data = self.process_input_data(test_data, processing_config)
                # train_data,test_data = np.stack(train_data.tolist()), np.stack(test_data.tolist())
                items.append([f'{column}', np.nan_to_num(train_data),
                              y_train_label, np.nan_to_num(test_data), y_test_label])

                # items.append([f'{column}_embedding', train_data,
                #             y_train_label, test_data, y_test_label])
        return items
 

    @computed_field
    @property
    def dataset_splitter(self) -> DatasetSplitter:
        return self.dataset_splitter_type()

    def set_dataframe(self, dataframe: pd.DataFrame):
        self.dataframe = dataframe

    @abc.abstractmethod
    def prep(self):
        """Prepare the dataset. This method should be overridden in subclasses."""
        

    def handle_mixins(self):
        """Handle mixin-specific logic."""
        if isinstance(self, TextDatasetMixin):
            self.process_text()
        # if isinstance(self, ImageDatasetMixin):
        #     self.process_image_data()
        # Add other mixin-specific logic here
        
    def load(self):
        """
        Load the dataset. If X_train, y_train, X_test, and y_test are already provided,
        skip deriving them. Otherwise, derive them from the dataframe and indices.
        """
        self.handle_mixins()  # Centralized mixin handling
        self.prep()  # Prepare the dataset

        if self.X_train is not None or self.y_train is not None or self.X_test is not None or self.y_test is not None:
            # Data is already provided, no need to calculate
            logging.info(
                "X_train, y_train, X_test, and y_test are already provided. Skipping calculation.")
            return
            # return self.X_train, self.X_test, self.y_train, self.y_test, self.train_indexes, self.test_indexes, self.train_idx_arr, self.val_idx_arr

        if self.index_path is None:
            raise Exception(
                "There is no index path. Please call split_dataset or provide indices.")

        if self.dataframe is None:
            raise Exception("There is no dataframe to derive data from.")

        try:
            train_idx_all, test_idx_all, train_idx_arr, val_idx_arr = self.__get_indexes__(
                self.index_path)
        except FileNotFoundError as e:
            raise FileNotFoundError(f"Index files not found: {e.filename}")
        
        # train = self.dataframe[self.dataframe.index.isin(train_idx_all)]
        # test = self.dataframe[self.dataframe.index.isin(test_idx_all)]
        columns = self.columns + [self.class_column] 
        train = self.dataframe.loc[self.dataframe.index.isin(train_idx_all), columns]
        test = self.dataframe.loc[self.dataframe.index.isin(test_idx_all), columns]
        X_train = train.drop(self.class_column, axis=1)
        X_train = train.drop(self.class_column, axis=1)
        y_train = train[self.class_column]
        X_test = test.drop(self.class_column, axis=1)
        y_test = test[self.class_column]

        self.X_train = np.array(X_train)
        # self.y_train = np.array(y_train)
        self.y_train = np.array(y_train.tolist())
        self.X_test = np.array(X_test)
        # self.y_test = np.array(y_test)
        self.y_test = np.array(y_test.tolist())

        self.train_indexes = X_train.index
        self.test_indexes = X_test.index
        self.train_idx_arr = train_idx_arr
        self.val_idx_arr = val_idx_arr

        # Dataframe to numpy array conversion

        # return self.X_train, self.X_test, self.y_train, self.y_test, self.train_indexes, self.test_indexes, self.train_idx_arr, self.val_idx_arr

    def __get_indexes__(self, path):
        train_index_path = path+'/train_indexes.txt'
        test_index_path = path+'/test_indexes.txt'
        train_fold_files = f'{path}/train_fold_*.txt'
        val_fold_files = f'{path}/validation_fold_*.txt'
        train_idx_arr = []
        val_idx_arr = []
        with open(train_index_path, 'r', encoding="utf8") as f:
            train_idx_all = [int(r) for r in f.readlines()]
        with open(test_index_path, 'r', encoding="utf8") as f:
            test_idx_all = [int(r) for r in f.readlines()]

        for filepath in glob.glob(train_fold_files):
            with open(filepath, 'r', encoding="utf8") as f:
                train_idx = [int(r) for r in f.readlines()]
                train_idx_arr.append(train_idx)
        for filepath in glob.glob(val_fold_files):
            with open(filepath, 'r', encoding="utf8") as f:
                val_idx = [int(r) for r in f.readlines()]
                val_idx_arr.append(val_idx)
        return train_idx_all, test_idx_all, train_idx_arr, val_idx_arr

    def __save_indexes__(self, path, filename, indexes):
        create_folder_if_not_exists(path)
        file_path = path + '/'+filename
        str_indexes = [str(index) for index in indexes]
        with open(file_path, 'w') as f:
            f.write('\n'.join(str_indexes))

    def split_dataset(self, save_indexes: bool = False):
        """
        Split the dataset into training and testing sets. This method is only available
        if a dataframe exists. If X_train, y_train, X_test, and y_test are already present,
        raise an error.
        """
        if self.X_train is not None or self.X_test is not None:
            raise Exception(
                "X_train and X_test are already present. Splitting is not allowed.")

        self.prep()
        if self.dataframe is None:
            raise Exception("There is no dataframe to split.")

        save_path = self.index_path

        X = self.dataframe.drop(self.class_column, axis=1)
        y = self.dataframe[self.class_column]

        X_train, X_test, y_train, y_test, X_train.index, X_test.index, train_idx_arr, val_idx_arr = self.dataset_splitter.split(
            X=X, y=y)
        self.X_train = np.array(X_train)
        self.X_test = np.array(X_test)
        self.y_train = np.array(y_train.tolist())
        self.y_test = np.array(y_test.tolist())
        self.train_indexes = X_train.index
        self.test_indexes = X_test.index
        self.train_idx_arr = train_idx_arr
        self.val_idx_arr = val_idx_arr

        if save_indexes:
            # train_pairs = [row['id1'].join(',').row['id2'] for index, row in X_train.iterrows()]
            self.__save_indexes__(
                save_path, 'train_indexes.txt', self.train_indexes.values)
            self.__save_indexes__(
                save_path, 'test_indexes.txt',  self.test_indexes.values)

            for i, (train_idx, val_idx) in enumerate(zip(train_idx_arr, val_idx_arr)):
                self.__save_indexes__(
                    save_path, f'train_fold_{i}.txt', train_idx)
                self.__save_indexes__(
                    save_path, f'validation_fold_{i}.txt', val_idx)

        # return X_train, X_test, y_train, y_test, folds


class TextDatasetMixin(BaseModel):
    embedding_dict: Dict[str, Any] | None = Field(
        default_factory=dict, description="Dictionary for embeddings")
    pooling_strategy: PoolingStrategy | None = None
    column_embedding_configs: Optional[List] = None
    vector_store_manager: BaseVectorStoreManager| None = None  # <-- NEW

    vector_db_persist_directory: Optional[str] = None
    vector_db_collection_name: Optional[str] = None
    _embedding_size: int

    @computed_field
    @property
    def embedding_size(self) -> int:
        return self._embedding_size

    class Config:
        arbitrary_types_allowed = True
 
    def __calculate_embedding_size(self):
        if not self.embedding_dict:
            raise ValueError("Embedding dictionary is not initialized, embedding size cannot be calculated.")
            
        key, value = next(iter(self.embedding_dict.items()))
        self._embedding_size = value[next(iter(value))][0].shape[0]

    def process_text(self):
        logging.info("Processing text data...")
 
      
        if not self.embedding_dict:
            if self.vector_store_manager is not None:
                self.embedding_dict = self.vector_store_manager.initialize_embedding_dict()
            else:
                raise ValueError("Either embedding_dict or vector_store_manager must be provided for text processing.")
        self.__calculate_embedding_size()
         


# class ImageDatasetMixin(BaseModel):
#     image_size: tuple[int, int] = Field(default=(224, 224))
#     augmentations: list[str] = Field(default_factory=list)

#     def process_image_data(self):
#         print(
#             f"Processing image data with size {self.image_size} and augmentations {self.augmentations}...")
